from typing import Any, List, Tuple, Optional

class Point:
    x: int
    y: int
    def __init__(self, x: int, y: int) -> None: ...
    def __repr__(self) -> str: ...

class Rect:
    x: int
    y: int
    width: int
    height: int
    def __init__(self, x: int, y: int, width: int, height: int) -> None: ...
    def __repr__(self) -> str: ...
    def get_position(self) -> Point: ...
    def get_size(self) -> Point: ...

class Match:
    rect: Rect
    label: str | None
    def __init__(self, rect: Rect, label: str | None = None) -> None: ...
    def __repr__(self) -> str: ...

class Color:
    r: int
    g: int
    b: int
    a: int
    def __init__(self, r: int, g: int, b: int, a: int = 255) -> None: ...
    def __repr__(self) -> str: ...

class Image:
    def __init__(self, width: int, height: int, channels: int = 3) -> None: ...
    def get_size(self) -> Point: ...
    def show(self, window_name: str = ...) -> None: ...
    def clone(self) -> 'Image': ...
    def crop(self, rect: Rect) -> None: ...
    def grayscale(self) -> None: ...
    def scale(self, x: float, y: float = ...) -> None: ...
    def rotate(self, angle: float) -> None: ...
    def flip(self, flip: int) -> None: ...
    def resize(self, size: Point) -> None: ...
    def draw_rect(self, rect: Rect, color: Optional[Color] = ..., thickness: int = 2) -> None: ...
    def draw_matches(self, matches: List[Match], color: Optional[Color] = ..., thickness: int = 2) -> None: ...
    def draw_line(self, start: Point, end: Point, color: Optional[Color] = ..., thickness: int = 2) -> None: ...
    def draw_text(self, text: str, pos: Point, color: Optional[Color] = ..., font_size: int = ..., thickness: int = ...) -> None: ...
    def draw_ellipse(self, center: Point, radius: int | Tuple[int, int], color: Optional[Color] = ..., thickness: int = 2, angle: float = 0) -> None: ...
    def draw_image(self, image: 'Image', pos: Point, alpha: float = 1.0) -> None: ...
    def invert(self) -> None: ...
    def brightness(self, value: float) -> None: ...
    def contrast(self, value: float = 1.0) -> None: ...
    def sharpen(self, value: float = 1.0) -> None: ...
    def blur(self, value: int = 3) -> None: ...
    def threshold(self, threshold: float = 128.0, maxValue: float = 255.0) -> None: ...
    def normalize(self, alpha: float = 0.0, beta: float = 255.0) -> None: ...
    def edge(self, threshold1: float = 100.0, threshold2: float = 200.0) -> None: ...
    def emboss(self) -> None: ...
    def split(self) -> List['Image']: ...
    def merge(self, channels: List['Image']) -> None: ...
    def convert(self, color_space: int) -> None: ...
    def range(self, lower: Color, upper: Color) -> None: ...
    def mask(self, mask: 'Image') -> None: ...

def load(path: str) -> str | Image | List[str|Image]: ...
def save(image: Image, path: str) -> None: ...
def capture(display_index: int = ..., rect: Rect = ...) -> Image: ...
def find_image(source: Image, search: Image, threshold: float = 0.8) -> List[Match]: ...
def find_text(source: Image, search: str, threshold: float = 0.0, text_level: int = ...) -> List[Match]: ...
def find_any(source: Image, *search: str | Image, threshold: float = 0.8, text_level: int = ...) -> List[Match]: ...
def find_all(source: Image, *search: str | Image, threshold: float = 0.8, text_level: int = ...) -> List[Match]: ...
def expect_any(*search: str | Image, interval: float = 1.0, timeout: float = -1.0, display_index: int = 0, threshold: float = 0.8, text_level: int = ...) -> List[Match]: ...
def expect_all(*search: str | Image, interval: float = 1.0, timeout: float = -1.0, display_index: int = 0, threshold: float = 0.8, text_level: int = ...) -> List[Match]: ...
def wait(seconds: float) -> None: ...
def mouse_move(pos: Any, display_index: int = ...) -> None: ...
def mouse_click(button: int = ..., count: int = ...) -> None: ...
def mouse_down(button: int = ...) -> None: ...
def mouse_up(button: int = ...) -> None: ...
def mouse_scroll(vertical: int, horizontal: int = ...) -> None: ...
def mouse_get_location() -> Tuple[Point, int]: ...
def mouse_get_display() -> int: ...
def type(text: str, wait: float = ...) -> None: ...
def key_click(key: int, count: int = ...) -> None: ...
def key_down(key: int) -> None: ...
def key_up(key: int) -> None: ...
def record(output_path: str, simplify: bool = ..., stop_key: int = ...) -> None: ...
def display_get_rect(display_index: int = ...) -> Rect: ...

# Constants
TEXT_BLOCK: int
TEXT_PARAGRAPH: int
TEXT_LINE: int
TEXT_WORD: int
TEXT_SYMBOL: int

DISPLAY_COUNT: int

KEY_META: int
KEY_BACKSPACE: int
KEY_TAB: int
KEY_ENTER: int
KEY_SHIFT: int
KEY_CTRL: int
KEY_ALT: int
KEY_PAUSE: int
KEY_CAPSLOCK: int
KEY_ESC: int
KEY_SPACE: int
KEY_PAGEUP: int
KEY_PAGEDOWN: int
KEY_END: int
KEY_HOME: int
KEY_LEFT: int
KEY_UP: int
KEY_RIGHT: int
KEY_DOWN: int
KEY_PRINTSCREEN: int
KEY_INSERT: int
KEY_DELETE: int
KEY_NUMLOCK: int
KEY_SCROLLLOCK: int
KEY_NUMPAD0: int
KEY_NUMPAD1: int
KEY_NUMPAD2: int
KEY_NUMPAD3: int
KEY_NUMPAD4: int
KEY_NUMPAD5: int
KEY_NUMPAD6: int
KEY_NUMPAD7: int
KEY_NUMPAD8: int
KEY_NUMPAD9: int
KEY_MULTIPLY: int
KEY_ADD: int
KEY_SEPARATOR: int
KEY_SUBTRACT: int
KEY_DECIMAL: int
KEY_DIVIDE: int
KEY_0: int
KEY_1: int
KEY_2: int
KEY_3: int
KEY_4: int
KEY_5: int
KEY_6: int
KEY_7: int
KEY_8: int
KEY_9: int
KEY_A: int
KEY_B: int
KEY_C: int
KEY_D: int
KEY_E: int
KEY_F: int
KEY_G: int
KEY_H: int
KEY_I: int
KEY_J: int
KEY_K: int
KEY_L: int
KEY_M: int
KEY_N: int
KEY_O: int
KEY_P: int
KEY_Q: int
KEY_R: int
KEY_S: int
KEY_T: int
KEY_U: int
KEY_V: int
KEY_W: int
KEY_X: int
KEY_Y: int
KEY_Z: int
KEY_F1: int
KEY_F2: int
KEY_F3: int
KEY_F4: int
KEY_F5: int
KEY_F6: int
KEY_F7: int
KEY_F8: int
KEY_F9: int
KEY_F10: int
KEY_F11: int
KEY_F12: int

BUTTON_LEFT: int
BUTTON_RIGHT: int
BUTTON_MIDDLE: int
BUTTON_X1: int
BUTTON_X2: int

SIMPLIFY_NONE: int
SIMPLIFY_MOVE: int
SIMPLIFY_MOUSE: int
SIMPLIFY_KEY: int
SIMPLIFY_TIME: int
SIMPLIFY_ALL: int

FLIP_VERTICAL: int
FLIP_HORIZONTAL: int
FLIP_BOTH: int

COLOR_SPACE_UNKNOWN: int
COLOR_SPACE_BGR: int
COLOR_SPACE_BGRA: int
COLOR_SPACE_RGB: int
COLOR_SPACE_RGBA: int
COLOR_SPACE_GRAY: int
COLOR_SPACE_HSV: int


import time
import threading
from pynput import mouse, keyboard
import chivel

# Map pynput key to Windows virtual-key code
def key_to_vk(key):
    # Handle special keys
    if hasattr(key, 'vk'):
        return key.vk
    if hasattr(key, 'value') and hasattr(key.value, 'vk'):
        return key.value.vk
    # Fallback for character keys
    try:
        return ord(key.char.upper())
    except Exception:
        return None
    
ACTION_NONE = 0
ACTION_MOUSE_MOVE = 1
ACTION_MOUSE_CLICK = 2
ACTION_MOUSE_DOWN = 3
ACTION_MOUSE_UP = 4
ACTION_MOUSE_SCROLL = 5
ACTION_KEY_CLICK = 6
ACTION_KEY_DOWN = 7
ACTION_KEY_UP = 8
ACTION_TYPE = 9

SIMPLIFY_MOVE = 1
SIMPLIFY_CLICK = 2
SIMPLIFY_TYPE = 4
SIMPLIFY_TIME = 8

TIME_CLICK = 0.5 # threshold for a click in seconds
TIME_TYPE = 1.0 # threshold for typing in seconds
TIME_PAUSE = 1.0 # time to wait between actions in seconds
TIME_ROUND = 0.250 # time to round to for actions in seconds

def record(output_path, simplify, stop_key=123):  # 123 is F12
    """
    Record mouse and keyboard actions until stop_key (virtual-key code) is pressed.
    Writes a Python script using the chivel library to replay the actions.
    """
    actions = []
    start_time = time.time()
    stop_flag = threading.Event()

    def now():
        return time.time() - start_time

    def on_move(x, y):
        display_index = chivel.mouse_get_display()
        display_rect = chivel.display_get_rect(display_index)
        actions.append((now(), ACTION_MOUSE_MOVE, (display_index, x - display_rect[0], y - display_rect[1])))

    def on_click(x, y, button, pressed):
        if pressed:
            actions.append((now(), ACTION_MOUSE_DOWN, (x, y, button.name)))
        else:
            actions.append((now(), ACTION_MOUSE_UP, (x, y, button.name)))

    def on_scroll(x, y, dx, dy):
        actions.append((now(), ACTION_MOUSE_SCROLL, (dx, dy)))

    def on_press(key):
        vk = key_to_vk(key)
        actions.append((now(), ACTION_KEY_DOWN, (vk, key.char if hasattr(key, 'char') else None)))
        if vk == stop_key:
            stop_flag.set()
            return False

    def on_release(key):
        vk = key_to_vk(key)
        actions.append((now(), ACTION_KEY_UP, (vk, key.char if hasattr(key, 'char') else None)))

    mouse_listener = mouse.Listener(on_move=on_move, on_click=on_click, on_scroll=on_scroll)
    keyboard_listener = keyboard.Listener(on_press=on_press, on_release=on_release)
    mouse_listener.start()
    keyboard_listener.start()

    stop_flag.wait()
    mouse_listener.stop()
    keyboard_listener.stop()

    if simplify > 0:
        simplify_move = (simplify & SIMPLIFY_MOVE) != 0
        simplify_click = (simplify & SIMPLIFY_CLICK) != 0
        simplify_type = (simplify & SIMPLIFY_TYPE) != 0
        simplify_time = (simplify & SIMPLIFY_TIME) != 0
        oldActions = actions
        actions = []
        actions.append((0, ACTION_NONE, ()))
        
        # handles the simplification of actions
        def add_action(t, action, params):
            oldT, oldAction, oldParams = actions[-1]
            if simplify_move and action == ACTION_MOUSE_MOVE and oldAction == ACTION_MOUSE_MOVE and params[0] == oldParams[0]:
                # if another move, extend the move
                actions.pop()
                add_action(t, ACTION_MOUSE_MOVE, (params[0], params[1], params[2]))
            elif simplify_click and action == ACTION_MOUSE_UP and oldAction == ACTION_MOUSE_DOWN and params[2] == oldParams[2] and t - oldT <= TIME_CLICK:
                # if mouse up after mouse down, replace with click
                actions.pop()
                add_action(t, ACTION_MOUSE_CLICK, (params[0], params[1], params[2], 1))
            elif simplify_click and action == ACTION_MOUSE_CLICK and oldAction == ACTION_MOUSE_CLICK and params[2] == oldParams[2] and t - oldT <= TIME_CLICK:
                # if another click, extend the click count
                actions.pop()
                add_action(t, ACTION_MOUSE_CLICK, (params[0], params[1], params[2], params[3] + 1))
            elif simplify_click and action == ACTION_KEY_UP and oldAction == ACTION_KEY_DOWN and params[0] == oldParams[0] and t - oldT <= TIME_CLICK:
                # if key up after key down, replace with click
                actions.pop()
                add_action(t, ACTION_KEY_CLICK, (params[0], params[1], 1))
            elif simplify_click and action == ACTION_KEY_CLICK and params[1] is not None and oldAction == ACTION_KEY_CLICK and params[0] == oldParams[0] and t - oldT <= TIME_CLICK:
                # if another key click, extend the click count if the same type, or start typing if different
                actions.pop()
                add_action(t, ACTION_KEY_CLICK, (params[0], params[1], oldParams[2] + 1))
            elif simplify_type and action == ACTION_KEY_CLICK and params[1] is not None and oldAction == ACTION_KEY_CLICK and params[0] != oldParams[0] and t - oldT <= TIME_CLICK:
                # if another key click, extend the click count if the same type, or start typing if different
                actions.pop()
                add_action(t, ACTION_TYPE, (oldParams[1] + params[1]))
            elif simplify_type and action == ACTION_KEY_CLICK and params[1] is not None and oldAction == ACTION_TYPE and t - oldT <= TIME_TYPE:
                # extend typing if within threshold
                actions.pop()
                add_action(t, ACTION_TYPE, (oldParams[0] + params[1]))
            else:
                # otherwise, add a new action
                actions.append((t, action, params))
        
        # re-add each action and simplify as needed
        for t, action, params in oldActions:
            add_action(t, action, params)

        # simplify the times
        if simplify_time:
            last_time = 0
            last_action = ACTION_NONE
            for i in range(len(actions)):
                t, action, params = actions[i]
                if action == ACTION_NONE:
                    continue
                if t - last_time > TIME_PAUSE:
                    # shorten the time if too long since last action
                    actions[i] = (last_time + TIME_PAUSE, action, params)
                elif last_action == ACTION_NONE or last_action == ACTION_MOUSE_MOVE:
                    # no time after mouse move or if no previous action
                    actions[i] = (round(last_time / TIME_ROUND) * TIME_ROUND, action, params)
                else:
                    # round the time to the nearest TIME_ROUND
                    dt = t - last_time
                    actions[i] = (last_time + (round(dt / TIME_ROUND) * TIME_ROUND), action, params)
                last_time = actions[i][0]
                last_action = action

        # remove any ACTION_NONE entries
        actions = [a for a in actions if a[1] != ACTION_NONE]

    with open(output_path, 'w', encoding='utf-8') as f:
        f.write("import chivel as c\n\n")
        f.write("def play():\n")
        last_time = 0
        for t, action, params in actions:
            if action == ACTION_NONE:
                continue
            wait = t - last_time - 0.001
            if wait >= 0.001:
                f.write(f"    c.wait({wait:.3f})\n")
            if action == ACTION_MOUSE_MOVE:
                f.write(f"    c.mouse_move({params[0]}, ({params[1]}, {params[2]}))\n")
            elif action == ACTION_MOUSE_DOWN:
                btn = 0 if params[2] == 'left' else 1 if params[2] == 'right' else 2
                f.write(f"    c.mouse_down({btn})\n")
            elif action == ACTION_MOUSE_UP:
                btn = 0 if params[2] == 'left' else 1 if params[2] == 'right' else 2
                f.write(f"    c.mouse_up({btn})\n")
            elif action == ACTION_MOUSE_CLICK:
                btn = 0 if params[2] == 'left' else 1 if params[2] == 'right' else 2
                f.write(f"    c.mouse_click({btn})\n")
            elif action == ACTION_MOUSE_SCROLL:
                f.write(f"    c.mouse_scroll({params[1]}, {params[0]})\n")
            elif action == ACTION_KEY_DOWN:
                if params[0] != stop_key:
                    f.write(f"    c.key_down({params[0]})\n")
            elif action == ACTION_KEY_UP:
                if params[0] != stop_key:
                    f.write(f"    c.key_up({params[0]})\n")
            elif action == ACTION_KEY_CLICK:
                if params[0] != stop_key:
                    f.write(f"    c.key_click({params[0]}, count={params[2]})\n")
            elif action == ACTION_TYPE:
                text = params[0].replace("'", "\\'").replace('\n', '\\n')
                f.write(f"    c.type('{text}', wait=0.05)\n")
            last_time = t
        f.write("\nif __name__ == '__main__':\n    play()\n")



from setuptools import setup

setup(
    name="chivel",
    version="0.6.0",
    description="For easily interfacing with the screen, just as a human would.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    # long_description="Chivel is a Python module that allows you to interact with the screen in a way that mimics human behavior. It provides functions for recording and playing back mouse and keyboard actions, making it useful for automation tasks.",
    author="Mitchell Talyat",
    packages=["chivel"],
    package_dir={"chivel": "chivel"},
    package_data={"chivel": [
        "chivel.pyd", 
        "chivel.pyi",
        "archive.dll",
        "bz2.dll",
        "gif.dll",
        "jpeg62.dll",
        "leptonica-1.85.0.dll",
        "libcrypto-3-x64.dll",
        "libcurl.dll",
        "liblzma.dll",
        "libpng16.dll",
        "libsharpyuv.dll",
        "libssl-3-x64.dll",
        "libwebp.dll",
        "libwebpdecoder.dll",
        "libwebpdemux.dll",
        "libwebpmux.dll",
        "libxml2.dll",
        "lz4.dll",
        "opencv_world4110.dll",
        "openjp2.dll",
        "tesseract55.dll",
        "tiff.dll",
        "turbojpeg.dll",
        "zlib1.dll",
        "zstd.dll",
        "tessdata/*"
    ]},
    include_package_data=True,
)